/*
 * Copyright 2012 Steven Gribble.
 *    [with contributions from cse333 students.]
 *
 *  This file is part of the UW CSE333 project sequence (333proj).
 *
 *  333proj is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  333proj is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 333proj.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdint.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <sys/select.h>
#include <string>
#include <iostream>
#include <sstream>
#include <fstream>
#include <stdexcept>

#include "./JSONParser.h"

#include "./test_suite.h"
#include "./test_jsonparser.h"

namespace hw2 {

std::string SlurpFile(const char *filename) {
  std::ifstream t(filename);
  std::stringstream buffer;
  buffer << t.rdbuf();
  return buffer.str();
}

TEST_F(Test_JSONParser, TESTJSONParserURLEncoder) {
  unsigned char test1[] = "Hello";
  unsigned char test2[] = " !*'(); :@&=+ $,/?#[] ";
  unsigned char encoded[] =
    "+%21%2a%27%28%29%3b+%3a%40%26%3d%2b+%24%2c%2f%3f%23%5b%5d+";

  ASSERT_STREQ((const char *) test1,
               URIEncode((const char *) test1).c_str());
  ASSERT_STREQ((const char *) encoded,
               URIEncode((const char *) test2).c_str());
}

TEST_F(Test_JSONParser, TestJSONParserUTFConversions) {
  unsigned char test1[] = { 0x6a }, test2[] = { 0xca, 0xaa },
                test3[] = { 0xed, 0xbb, 0xB3 };
  char test4[] = "\\u007F",
    test5[] = "\\u07FF",
    test6[] = "\\uFFFF";
  char outbuf[7];
  unsigned char outbytes[4];
  int res;

  res = UTF8ToSlashUEncoded(test1, outbuf);
  ASSERT_STREQ("\\u006A", outbuf);
  ASSERT_EQ(1, res);

  res = UTF8ToSlashUEncoded(test2, outbuf);
  ASSERT_STREQ("\\u02AA", outbuf);
  ASSERT_EQ(2, res);

  res = UTF8ToSlashUEncoded(test3, outbuf);
  ASSERT_STREQ("\\uDEF3", outbuf);
  ASSERT_EQ(3, res);

  SlashUEncodedToUTF8(test4, outbytes);
  ASSERT_EQ(0x7F, outbytes[0]);
  ASSERT_EQ(0x00, outbytes[1]);

  SlashUEncodedToUTF8(test5, outbytes);
  ASSERT_EQ(0xDF, outbytes[0]);
  ASSERT_EQ(0xBF, outbytes[1]);
  ASSERT_EQ(0x00, outbytes[2]);

  SlashUEncodedToUTF8(test6, outbytes);
  ASSERT_EQ(0xEF, outbytes[0]);
  ASSERT_EQ(0xBF, outbytes[1]);
  ASSERT_EQ(0xBF, outbytes[2]);
  ASSERT_EQ(0x00, outbytes[3]);

  HW2Addpoints(5);
}

TEST_F(Test_JSONParser, TestJSONParserEncodings) {
  // Test JSONEscapeString
  unsigned char unescaped_source[] = "\"test\" \xed\xbb\xb3 \xca\xaa good\r\n";
  std::string escaped_dest = JSONEscapeString(unescaped_source);
  ASSERT_STREQ("\\\"test\\\" \\uDEF3 \\u02AA good\\r\\n",
               reinterpret_cast<const char *>(escaped_dest.c_str()));

  // Test JSONUnescapeString
  std::string unescaped_dest =
    JSONUnescapeString(reinterpret_cast<unsigned char *>(
                        const_cast<char *>(escaped_dest.c_str())));
  ASSERT_STREQ(reinterpret_cast<const char *>(unescaped_source),
               unescaped_dest.c_str());
  HW2Addpoints(5);
}

TEST_F(Test_JSONParser, TestJSONParserGenParse) {
  try {
    // Simple JSON string.
    char simplestr[] =
      "{ \"a1\": true, \"name\": \"val\", \"name2\": \"val\\\"2\\\"\", "
      "\"name3\": { \"name4\": \"val4\", \"name5\": [ 3, 4, 5, "
      "{ \"name6\": true } ] } }";

    // Parse it.
    JSONValue *jvp = JSONValue::JSONParse(simplestr);

    // Convert back to string, make sure is the same.
    std::stringstream ss;
    ss << *jvp;
    ASSERT_STREQ(simplestr, ss.str().c_str());
    delete jvp;

    // Do the same JSON string, but wonky whitespacing, and make sure the
    // generated string matches simplestr[].
    char simplestr2[] =
      "  {\"a1\":true,  \"name\": \"val\" , \"name2\" :\"val\\\"2\\\"\", "
      "\"name3\": { \"name4\":\"val4\",\"name5\":[3,4,5,{\"name6\":true}]}}  ";
    jvp = JSONValue::JSONParse(simplestr2);
    std::stringstream ss22;
    ss22 << *jvp;
    ASSERT_STREQ(simplestr, ss22.str().c_str());

    // Do a little data structure traversal.
    JSONValue *name = ((reinterpret_cast<JSONObject *>(jvp))->Get("name"));
    ASSERT_STREQ("val", ((reinterpret_cast<JSONString *>(name))->str_.c_str()));

    // Clean up the parsed JSON tree.
    delete jvp;

    // Complex cases;  all we'll do is make sure we can parse.
    std::string complex1 = SlurpFile("test_json_files/1.json");
    std::string complex2 = SlurpFile("test_json_files/2.json");
    std::string complex3 = SlurpFile("test_json_files/3.json");

    // Parse the first.
    jvp = JSONValue::JSONParse(complex1);
    std::stringstream ss2;
    ss2 << *jvp;
    delete jvp;

    // Parse the second.
    jvp = JSONValue::JSONParse(complex2);
    std::stringstream ss3;
    ss3 << *jvp;
    delete jvp;

    // Parse the third.
    jvp = JSONValue::JSONParse(complex3);

    // Generate the string from the third, and parse that.
    std::stringstream ss4;
    ss4 << *jvp;
    delete jvp;
    jvp = JSONValue::JSONParse(ss4.str());
    std::stringstream ss5;
    ss5 << *jvp;
    delete jvp;

    // Make sure the parsed third and latest are the same.
    ASSERT_EQ(ss4.str(), ss5.str());
  } catch(std::runtime_error e) {
    std::cerr << "caught '" << e.what() << "'\n";
    FAIL();
  }
}

}  // namespace hw2
