/*
 * Copyright 2012 Steven Gribble
 *
 *  This file is part of the UW CSE333 project sequence (333proj12).
 *
 *  333proj12 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published
 *  by the Free Software Foundation, either version 3 of the License,
 *  or (at your option) any later version.
 *
 *  333proj12 is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with 333proj12.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <boost/algorithm/string.hpp>

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <vector>
#include <map>
#include <algorithm>
#include <sstream>

#include "./TwitterAPI.h"
#include "./JSONParser.h"
#include "./Utils.h"

using std::cout;
using std::endl;
using std::map;
using std::pair;
using std::string;
using std::stringstream;
using std::vector;

// Prompt the user to type in a query, and slurp it into a string.
string GetQuery(void);

// Given a query string, does a twitter query to fetch a list of words
// associated with the query string.  HTML encodes the words to make
// them safe to embed in a javascript string.
vector<string> GetHTMLEncodedWordList(string querystring);

// Given a vector of words, produce a sorted vector of <string,int>
// pairs.
vector<pair<int, string> > CountAndSort(vector<string> &words);

// Given the sorted vector, produce HTML for the word cloud.
std::string GenerateHTML(vector<pair<int, string> > &countwords);

bool mysort(pair<int, string> x, pair<int, string> y) {
  return x.first > y.first;
}

int main(int argc, char **argv) {
  // Loop, reading in a query and processing it.
  while (1) {
    string nextquery = GetQuery();
    if (nextquery.size() == 0)
      break;

    // Get the list of words from the query reply.
    vector<string> words = GetHTMLEncodedWordList(nextquery);

    // Produce a frequency count of the vectors; do it with a map from
    // word to count.
    vector<pair<int, string> > countwords = CountAndSort(words);

    // Generate the HTML.
    std::string html = GenerateHTML(countwords);

    // Write the data out to a tmp file.
    FILE *f = fopen("cloud.html", "w");
    fwrite(html.c_str(), html.size(), 1, f);
    fclose(f);

    // Open in firefox, delete the file.
    std::string ffcommand = "firefox cloud.html";
    system(ffcommand.c_str());
    unlink("cloud.html");
  }
  return EXIT_SUCCESS;
}

vector<string> GetHTMLEncodedWordList(string querystring) {
  vector<string> results;

  // Do the query, return empty vector on failure.
  // STEP 1

  // Loop through the results, pulling out the "text" field from and
  // splitting individual words out of them.  Do some sanitization on
  // the words; convert them to lower case and remove punctuation.
  // Insert each of the santized words into the "results" vector, and
  // return that.
  // STEP 2

  return results;
}

vector<pair<int, string> > CountAndSort(vector<string> &words) {
  // Construct a word frequency count in a map.
  map<string, int> frequency;
  for (uint64_t i = 0U; i < words.size(); i++) {
    int count = frequency[words[i]];
    frequency[words[i]]  = ++count;
  }

  // Construct the vector of count / word pairs.
  vector<pair<int, string> > countwords;
  for (map<string, int>::iterator it = frequency.begin();
       it != frequency.end(); it++) {
    pair<int, string> nextp(static_cast<int>(sqrt(it->second)), it->first);
    countwords.push_back(nextp);
  }

  // Sort and return the vector.
  sort(countwords.begin(), countwords.end(), &mysort);
  return countwords;
}

std::string GenerateHTML(vector<pair<int, string> > &countwords) {
  stringstream ss;

  // Generate the header.
  ss << "<html><head>\n";
  // FINISH THE HEADER GENERATION
  // STEP 3

  // Generate the top of the body.
  // STEP 4

  // Generate the word list.
  // STEP 5

  // Generate the bottom of the HTML.
  // STEP 6

  return ss.str();
}

string GetQuery(void) {
  size_t n = 0;
  char *lineptr = NULL;

  // Prompt the user to enter a query.
  cout << "Enter query:" << endl;

  // Read the next query, strip the newline.  Use getline()
  // ("man getline").  If getline() returns -1, be sure to
  // free() the lineptr if it is non-NULL, and return an empty
  // string to indicate we've hit the end of file.
  int res = getline(&lineptr, &n, stdin);
  if (res == -1) {
    if (lineptr != NULL)
      free(lineptr);
    return "";
  }

  // Create the string to return.
  std::string retstr(lineptr);

  // Trim whitespace.
  boost::trim(retstr);

  // free() up any memory malloc'ed by getline() and returned via lineptr.
  if (lineptr != NULL) {
    free(lineptr);
    lineptr = NULL;
  }
  return retstr;
}
